classdef RTree < handle

  ## -*- texinfo -*-
  ## @deftp {Class} RTree
  ##
  ## An Octave native extension implementing the R-tree spatial index
  ## of Guttman-Green. The code is an embedded version of librtree.
  ##
  ## Given a set or rectangles (or higher dimensional cuboids) and their
  ## associated ids, one can build an RTree using the @code{csv_read} class
  ## method or repeated calls to the @code{add_rect} instance method. The
  ## former is more performant since it avoids routing the rectangles
  ## through the Octave-C interface.
  ##
  ## Once the RTree instance is created one can make spatial queries
  ## against it with the @code{search method}; one passes a rectangle to
  ## the method and it applies a user-supplied callback function to the
  ## ids of all of the input rectangles which intersect it.
  ##
  ## It may be convenient to serialise the RTree for faster loading, the
  ## library implements a custom binary format, BSRT (binary serialised
  ## R-tree) which can be written by @code{bsrt_write} and read with the
  ## @code{bsrt_read} class method.  One can also serialise to JSON, but
  ## this results in a much larger file (a factor of ten) and so
  ## correspondingly slow to read/write. Useful, nevertheless, for
  ## debugging.
  ##
  ## All rectangles used in the library are simply arrays of floats,
  ## the lower value for each dimension, followed by the upper value for
  ## each dimension. Thus
  ##
  ## @example
  ## [0, 0, 1, 2]
  ## @end example
  ##
  ## @noindent
  ## is the rectangle with 0 ≤ x ≤ 1 and 0 ≤ y ≤ 2. Upper and lower values
  ## may coincide (to create a line segment in 2-space, for example) but a
  ## lower value larger than the upper value will cause an error.
  ##
  ## It is anticipated that the ids passed to the library will be used as
  ## an index for application specific data to which this rectangle relates
  ## (its location in an array, or a DB id) but this is entirely at the
  ## discretion of the caller, the library makes no use of the value,
  ## treating it as payload. In particular, the value may be non-unique and
  ## may be zero. One should note that the id type used internally by the
  ## library is determined at compile-time (with the @code{RTREE_ID_TYPE}
  ## variable) and by default this is a 64-bit unsigned integer.
  ##
  ## @end deftp
  ##
  ## @deftypefn {RTree} {@var{obj} =} RTree @
  ##    (@var{dim}, @var{split}, @var{np})
  ##
  ## Create a new (empty) RTree instance.
  ##
  ## @table @var
  ## @item dim
  ## The dimension of the tree to create, an integer greater than 0;
  ## @item split
  ## A string indicating the splitting strategy, one of @qcode{"linear"},
  ## @qcode{"quadratic"} or @qcode{"greene"}, default @qcode{"quadratic"}.
  ## The linear strategy is faster to build, the quadratic and Greene
  ## strategies produce better-quality R-trees which are faster to query.
  ## @item np
  ## The number of nodes-per-page, default 0.  This value can affect
  ## performance quite dramatically, particularly build time. A value
  ## which is too large would result in an infeasible branching factor
  ## for the R-tree and will cause the function to error.  A value of
  ## zero is permitted and the default; in this case the function will
  ## choose a good value based on heuristics. You may get better
  ## performance for your use-case by manual experimentation, but zero
  ## is a good place to start.
  ## @end table
  ##
  ## @seealso{RTree.csv_read, RTree.add_rect}
  ##
  ## @end deftypefn

  properties (Access = private)
    state
  endproperties

  methods (Static = true)

    function obj = RTree(dim, split = 'quadratic', node_page = 0)
      if nargin > 0
        dim = RTree.validate_dim(dim);
        split = RTree.validate_split(split);
        node_page = RTree.validate_node_page(node_page);
        obj.state = RTree_dispatch('constructor', dim, split, node_page);
      endif
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{str} =} bugreport ()
    ##
    ## The email address for bug reports.
    ##
    ## @end deftypefn

    function val = bugreport
      val = RTree_dispatch('bugreport');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{obj} =} bsrt_read (@var{io_arg})
    ## Create a new RTree instance from BSRT (binary serialised R-tree)
    ## path or readable binary stream @var{io_arg}.
    ##
    ## Example:
    ##
    ## @example
    ## rtree = RTree.bsrt_read ("file.bsrt");
    ## @end example
    ##
    ## @seealso{RTree.from_bsrt, RTree.bsrt_write}
    ##
    ## @end deftypefn

    function obj = bsrt_read(io_arg)
      if nargin < 1
        error('RTree:args-number', 'argument IO required')
      endif
      obj = RTree();
      obj.state = RTree.io_read('bsrt_read', io_arg, 'rb');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{obj} =} from_bsrt (@var{bsrt})
    ## Create (deserialise) an RTree instance from BSRT string
    ## @var{bsrt}.
    ##
    ## @seealso{RTree.to_bsrt, RTree.bsrt_read}
    ##
    ## @end deftypefn

    function obj = from_bsrt(bsrt)
      if nargin < 1
        error('RTree:args-number', 'argument BSRT required')
      endif
      obj = RTree();
      obj.state = RTree_dispatch('from_bsrt', bsrt);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{obj} =} csv_read @
    ## (@var{io_arg}, @var{dim}, @var{split}, @var{np})
    ## Create a new RTree instance from CSV path or stream.
    ##
    ## @table @var
    ## @item io_arg
    ## A path or readable text-stream of CSV data;
    ## @item dim
    ## The dimension of the tree to construct;
    ## @item split
    ## A string indicating the splitting strategy, one of @qcode{"linear"},
    ## @qcode{"quadratic"} or @qcode{"greene"}, default @qcode{"quadratic"};
    ## @item np
    ## The number of nodes-per-page, or 0 for heuristic choice, default 0
    ## @end table
    ##
    ## The CSV file (without header) should have the id in the first column,
    ## then twice as many floats as the dimension. Extra columns may be
    ## present and will be ignored (this useful feature is the reason that
    ## the dimension @var{dim} is a required argument).
    ##
    ## @end deftypefn

    function obj = csv_read(io_arg, dim,
                            split = 'quadratic',
                            node_page = 0)
      if nargin < 2
        error('RTree:args-number', 'argument DIM required')
      endif

      dim = RTree.validate_dim(dim);
      split = RTree.validate_split(split);
      node_page = RTree.validate_node_page(node_page);

      obj = RTree();
      obj.state = RTree.io_read('csv_read', io_arg, 'r', dim, split, node_page);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{obj} =} json_read (@var{io_arg})
    ## Create (deserialise) an RTree instance from JSON path or
    ## stream @var{io_arg}.
    ##
    ## Example:
    ##
    ## @example
    ## rtree = RTree.json_read ("file.json");
    ## @end example
    ##
    ## @seealso{RTree.to_json, RTree.json_write}
    ##
    ## @end deftypefn

    function obj = json_read(io_arg)
      if nargin < 1
        error('RTree:args-number', 'argument IO required')
      endif
      obj = RTree();
      obj.state = RTree.io_read('json_read', io_arg, 'rb');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{obj} =} from_json (@var{json})
    ## Create (deserialise) an RTree instance from JSON string
    ## @var{json}.
    ##
    ## @seealso{RTree.json_read, RTree.bsrt_read}
    ##
    ## @end deftypefn

    function obj = from_json(json)
      if nargin < 1
        error('RTree:args-number', 'argument JSON required')
      endif
      obj = RTree();
      obj.state = RTree_dispatch('from_json', json);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{str} =} name ()
    ## The name of the wrapped C library.
    ##
    ## @end deftypefn

    function val = name
      val = RTree_dispatch('name');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{str} =} url ()
    ## The URL of the package's homepage.
    ##
    ## @end deftypefn

    function val = url
      val = RTree_dispatch('url');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{a} =} version ()
    ## The project's version as a triple of integers.
    ##
    ## @end deftypefn

    function val = version
      version_string = RTree_dispatch('version');
      [vald, count] = sscanf(version_string, '%i.%i.%i', 3);
      if count != 3
        error('RTree:BUG', 'returned %i components', count);
      endif
      val = uint16(vald);
    endfunction

  endmethods

  methods

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {} add_rect (@var{id}, @var{rect})
    ##
    ## Add a rectangle @var{rect} to the RTree instance.
    ##
    ## @end deftypefn

    function add_rect(obj, id, rect)
      if nargin < 3
        error('RTree:args-number', 'expected 2 arguments');
      endif
      id = RTree.validate_id(id);
      rect = RTree.validate_rect(rect);
      RTree_dispatch('add_rect', obj.state, id, rect);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} branch_size ()
    ##
    ## The size @var{n} in bytes of a branch.
    ##
    ## @end deftypefn

    function val = branch_size(obj)
      val = RTree_dispatch('branch_size', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} branching_factor ()
    ##
    ## The number @var{n} of branches from each node.
    ##
    ## @end deftypefn

    function val = branching_factor(obj)
      val = RTree_dispatch('branching_factor', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {} bsrt_write (@var{io_arg})
    ##
    ## Serialise to BSRT (binary serialised R-tree) stream or
    ## path @var{io_arg}.
    ##
    ## @end deftypefn

    function bsrt_write(obj, io_arg)
      io_write(obj, 'bsrt_write', io_arg, 'wb');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{bsrt} = } to_bsrt ()
    ## Serialise to BSRT string.
    ##
    ## Example:
    ##
    ## @example
    ## bsrt = rtree.to_bsrt ();
    ## @end example
    ##
    ## @seealso{RTree.bsrt_write}
    ##
    ## @end deftypefn

    function bsrt = to_bsrt(obj)
      bsrt = RTree_dispatch('to_bsrt', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{obj} =} clone ()
    ##
    ## Create a deep copy.
    ##
    ## @end deftypefn

    function val = clone(obj)
      val = RTree();
      val.state = RTree_dispatch('clone', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {} csv_write (@var{io_arg})
    ##
    ## Serialise to CSV stream or path @var{io_arg}.
    ##
    ## @end deftypefn

    function csv_write(obj, io_arg)
      io_write(obj, 'csv_write', io_arg, 'wb');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{csv} = } to_csv ()
    ## Serialise to CSV string.
    ##
    ## Example:
    ##
    ## @example
    ## csv = rtree.to_csv ();
    ## @end example
    ##
    ## @seealso{RTree.csv_write}
    ##
    ## @end deftypefn

    function csv = to_csv(obj)
      csv = RTree_dispatch('to_csv', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} dim ()
    ##
    ## The dimension @var{n} of the R-tree.
    ##
    ## @end deftypefn

    function val = dim(obj)
      val = RTree_dispatch('dim', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{b} =} empty ()
    ##
    ## Whether or not the tree is empty (has no rectangles)
    ##
    ## @end deftypefn

    function val = empty(obj)
      val = RTree_dispatch('empty', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{rect} =} envelope ()
    ##
    ## Return the envelope, the minimal rectange which contains all of
    ## the rectangles so-far inserted into the RTree,  Will raise if
    ## the rtree is empty.
    ##
    ## @end deftypefn

    function val = envelope(obj)
      val = RTree_dispatch('envelope', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{bool} =} eq (@var{other})
    ##
    ## Equality of RTrees.
    ##
    ## As is usual with Octave, this method is called when one uses
    ## the infix equality operator, i.e., @code{A == B} is the same
    ## as @code{A.eq (B)}.
    ##
    ## This is a rather strict equality, not only must the tree have
    ## the same rectangles, they must be in the same order. Certainly
    ## @code{clone} will produce an instance which is equal in this
    ## sense.
    ##
    ## @seealso{RTree.ne, RTree.clone}
    ##
    ## @end deftypefn

    function val = eq(obj, other)
      val = RTree_dispatch('identical', obj.state, other.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} height ()
    ##
    ## The height of the tree in the usual mathematical sense.
    ##
    ## @end deftypefn

    function val = height(obj)
      val = RTree_dispatch('height', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {} json_write (io_arg)
    ## Serialise to JSON stream or path @var{io_arg}.
    ##
    ## Example:
    ##
    ## @example
    ## rtree.json_write ('file.json');
    ## @end example
    ##
    ## @seealso{RTree.json_read, RTree.bsrt_read}
    ##
    ## @end deftypefn

    function json_write(obj, io_arg)
      io_write(obj, 'json_write', io_arg, 'w');
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{json} = } to_json ()
    ## Serialise to JSON string.
    ##
    ## Example:
    ##
    ## @example
    ## json = rtree.to_json ();
    ## @end example
    ##
    ## @seealso{RTree.json_write}
    ##
    ## @end deftypefn

    function json = to_json(obj)
      json = RTree_dispatch('to_json', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{bool} =} ne (@var{other})
    ##
    ## Inequality of RTrees.
    ##
    ## As is usual with Octave, this method is called when one uses
    ## the infix inequality operator, i.e., @code{A != B} is the same
    ## as @code{A.ne (B)}.
    ##
    ## @seealso{RTree.eq}
    ##
    ## @end deftypefn

    function val = ne(obj, other)
      val = ! eq(obj, other);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} node_size ()
    ##
    ## The size @var{n} in bytes of a node.
    ##
    ## @end deftypefn

    function val = node_size(obj)
      val = RTree_dispatch('node_size', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} page_size ()
    ##
    ## The size @var{n} in bytes of a page of memory.
    ##
    ## @end deftypefn

    function val = page_size(obj)
      val = RTree_dispatch('page_size', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {} postscript @
    ## (@var{io_arg}, @var{style}, @var{axis}, @var{extent}, @var{margin})
    ##
    ## Create a PostScript plot of the RTree instance.
    ##
    ## @table @var
    ## @item io_arg
    ## A path, or writable text stream (as returned by @code{fopen});
    ## @item style
    ## An instance of the @code{RTreeStyle} class describing the fill
    ## colour and stroke width and colour for each level of the tree.
    ## @item axis
    ## The @var{axis} to which the @var{extent} variable is to be applied,
    ## one of @qcode{"height"} or @qcode{"width"} (default @qcode{"width"});
    ## @item extent
    ## The size of the output plot on the specified @var{axis} in units of
    ## PostScript point, 1/72 inch, (default @code{216} is 3 inches);
    ## @item margin
    ## Extra space around the plot in units of PostScript point, (default 0).
    ## @end table
    ##
    ## @seealso{RTreeStyle}
    ##
    ## @end deftypefn

    function postscript(obj, io_arg, style,
                        axis = 'width',
                        extent = 216,
                        margin = 0)
      if nargin < 3
        error('RTree:args-number', 'at least 2 arguments')
      endif
      io_write(obj, 'postscript', io_arg, 'w', style.state,
               axis, extent, margin);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} rect_size ()
    ##
    ## The size @var{n} in bytes of a rectangle.
    ##
    ## @end deftypefn

    function val = rect_size(obj)
      val = RTree_dispatch('rect_size', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{ctx} = } search @
    ## (@var{rect}, @var{f}, @var{ctx})
    ##
    ## Search the RTree for intersecting rectangles.
    ##
    ## @table @var
    ## @item rect
    ## The search rectangle;
    ## @item f
    ## A user-defined function which takes as argument  the @var{id} of
    ## found intersecting rectangle and the context @var{ctx} passed as
    ## the final argument, it should return this (usually modified)
    ## context;
    ## @item ctx
    ## Any object, this will be passed as the second input argument
    ## to the function @var{f} and should then be returned.
    ## @end table
    ##
    ## Examples:
    ##
    ## Count the number of intersecting rectangles:
    ##
    ## @example
    ## function val = f (id, count)
    ##   val = count + 1;
    ## endfunction
    ##
    ## count = rtree.search (rect, @@f, 0);
    ## @end example
    ##
    ## Return all ids of intersecting rectangles:
    ##
    ## @example
    ## function val = f (id, ids)
    ##   val = [ids, id];
    ## endfunction
    ##
    ## ids = rtree.search (rect, @@f, []);
    ## @end example
    ##
    ## Note that this sort of "accumulation" function will return
    ## results for all of the rectangles given a sufficiently large
    ## search rectangle; thus untrusted (user-supplied) search
    ## rectangles may cause excessive resource usage.
    ##
    ## @end deftypefn

    function val = search(obj, rect, f, context)
      if nargin < 4
        error('RTree:args-number', '3 arguments required');
      endif
      rect = RTree.validate_rect(rect);
      val = RTree_dispatch('search', obj.state, rect, f, context);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{n} =} size ()
    ##
    ## The total bytes @var{n} allocated for the instance.
    ##
    ## This method traverses the entire tree summing contributions
    ## for each node (rather than maintaining a running count).
    ## Performance-minded users may wish to cache this value
    ## (invalidating the cache when calling @code{add_rect} of course).
    ##
    ## @end deftypefn

    function val = size(obj)
      val = RTree_dispatch('size', obj.state);
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {} update (@var{f}, @var{ctx})
    ##
    ## Update the RTree instance. Modifies the rectangles in-place
    ## without changing the tree structure. Provided that the changes
    ## are small, the search efficiency should be close to that of
    ## freshly built RTree.
    ##
    ## @table @var
    ## @item f
    ## A user-defined function which takes as arguments a rectangle
    ## and the context @var{ctx} passed as the final argument, it should
    ## return the (usually modified) context;
    ## @item ctx
    ## Any object, this will be passed as the second input argument
    ## to the function @var{f}.  If that function does not access
    ## the context, then the @code{update} method can omit it.
    ## @end table
    ##
    ## Example:
    ##
    ## A function which shifts all rectangles by 1 on each axis:
    ##
    ## @example
    ## function val = f (rect, context)
    ##   val = rect + 1;
    ## endfunction
    ## @end example
    ##
    ## called as @code{rtree.update (@@f)}.
    ##
    ## Note that in the C library, the implementation (via a C callback
    ## function) is much faster than rebuilding the R-tree; in this
    ## Octave interface the callback must convert C floats to Octave,
    ## modify them and then convert the returned Octave floats to C; so
    ## we would expect that it loses much of its competitive advantage
    ## when compared to an R-tree rebuild.
    ##
    ## @end deftypefn

    function update(obj, f, context = 0)
      if nargin < 2
        error('RTree:args-number', 'at least 1 argument')
      endif
      err = RTree_dispatch('update', obj.state, f, context);
      if err != 0
        error('RTree:fail', 'update failed')
      endif
    endfunction

    ## -*- texinfo -*-
    ## @deftypefn {RTree} {@var{v} =} unit_sphere_volume ()
    ##
    ## The volume @var{v} of the unit sphere in the instance's
    ## dimension.
    ##
    ## @end deftypefn

    function val = unit_sphere_volume(obj)
      val = RTree_dispatch('unit_sphere_volume', obj.state);
    endfunction

  endmethods

  methods (Static = true, Access = private)

    function val = io_read(method, io_arg, mode, varargin)
      if is_valid_file_id(io_arg)
        val = RTree_dispatch(method, io_arg, varargin{:});
      elseif ischar(io_arg)
        io = fopen(io_arg, mode);
        try
          val = RTree_dispatch(method, io, varargin{:});
        catch
          fclose(io);
          rethrow(lasterror);
        end_try_catch
        fclose(io);
      else
        error('RTree:args-type', 'expected string or file descriptor');
      endif
    endfunction

    function val = validate_dim(dim)
      if ! isnumeric(dim)
        error('RTree:args-type', 'DIM should be numeric')
      endif
      if ! isscalar(dim)
        error('RTree:args-type', 'DIM should be scalar')
      endif
      if dim < 1
        error('RTree:args-invalid', 'DIM > 0 required')
      endif
      val = uint64(dim);
    endfunction

    function val = validate_id(id)
      if ! isnumeric(id)
        error('RTree:args-type', 'ID should be numeric');
      endif
      if ! isscalar(id)
        error('RTree:args-type', 'ID should be scalar');
      endif
      val = uint64(id);
    endfunction

    function val = validate_node_page(node_page)
      if ! isnumeric(node_page)
        error('RTree:args-type', 'NODE_PAGE should be numeric')
      endif
      if ! isscalar(node_page)
        error('RTree:args-type', 'NODE_PAGE should be scalar')
      endif
      if node_page < 0
        error('RTree:args-invalid', 'NODE_PAGE >= 0 required')
      endif
      val = uint32(node_page);
    endfunction

    function val = validate_rect(rect)
      if ! isnumeric(rect)
        error('RTree:args-type', 'RECT should be numeric');
      endif
      val = rect;
    endfunction

    function val = validate_split(split)
      if ! ischar(split)
        error('RTree:args-invalid', 'SPLIT should be string')
      endif
      val = split;
    endfunction

  endmethods

  methods (Access = private)

    function io_write(obj, method, io_arg, mode, varargin)
      if is_valid_file_id(io_arg)
        RTree_dispatch(method, obj.state, io_arg, varargin{:});
      elseif ischar(io_arg)
        io = fopen(io_arg, mode);
        try
          RTree_dispatch(method, obj.state, io, varargin{:});
        catch
          fclose(io);
          rethrow(lasterror);
        end_try_catch
        fclose(io);
      else
        error('RTree:args-type', 'expected string or file descriptor');
      endif
    endfunction

  endmethods

endclassdef
