// -*- c++ -*-

#ifndef RTREEOCT_H
#define RTREEOCT_H

#include <octave/oct.h>
#include <octave/parse.h>

#include "RTreeC.h"
#include "RTreeStyleOct.h"

class RTreeOct : public octave_base_value
{
public:

  // constructors

  RTreeOct()
    : pRTreeC(new RTreeC()) {}
  RTreeOct(RTreeC *rtc)
    : pRTreeC(rtc) {}
  RTreeOct(std::size_t dim,
           std::string split,
           std::uint32_t node_page)
    : pRTreeC(new RTreeC(dim, split, node_page)) {}

  // static (class) methods

  static octave_value constructor(octave_value,
                                  octave_value,
                                  octave_value);

  static octave_value bsrt_read(octave_value);
  static octave_value csv_read(octave_value,
                               octave_value,
                               octave_value,
                               octave_value);
  static octave_value json_read(octave_value);

  static octave_value from_bsrt(octave_value);
  static octave_value from_json(octave_value);

  static const octave_value bugreport(void)
  {
    return RTreeC::bugreport;
  }

  static const octave_value name(void)
  {
    return RTreeC::name;
  }

  static const octave_value url(void)
  {
    return RTreeC::url;
  }

  static const octave_value version(void)
  {
    return RTreeC::version;
  }

  // instance methods

  void add_rect(octave_value id_ov, octave_value rect_ov)
  {
    rtree_id_t id = id_ov.int_value();
    ColumnVector cv = rect_ov.column_vector_value();
    std::size_t n = cv.numel();
    std::vector<rtree_coord_t> rect(n);

    for (std::size_t i = 0 ; i < n ; i++)
      rect.at(i) = cv(i);

    pRTreeC->add_rect(id, rect);
  }

  octave_value branching_factor(void)
  {
    return octave_uint64(pRTreeC->branching_factor());
  }

  octave_value branch_size(void)
  {
    return octave_uint64(pRTreeC->branch_size());
  }

  void bsrt_write(octave_value fd_ov)
  {
    int fd = fd_ov.int_value();
    pRTreeC->bsrt_write(fd);
  }

  octave_value bytes(void)
  {
    return octave_uint64(pRTreeC->bytes());
  }

  octave_value clone(void)
  {
    return new RTreeOct(pRTreeC->clone());
  }

  void csv_write(octave_value fd_ov)
  {
    int fd = fd_ov.int_value();
    pRTreeC->csv_write(fd);
  }

  octave_value dim(void)
  {
    return octave_uint64(pRTreeC->dim());
  }

  octave_value empty(void)
  {
    return pRTreeC->empty();
  }

  octave_value envelope(void)
  {
    std::vector<double> u = pRTreeC->envelope();
    std::size_t n = u.size();
    Array<double> v(dim_vector(1, n), 0);

    for (std::size_t i = 0 ; i < n ; i++)
      v(i) = u[i];

    return v;
  }

  octave_value height(void)
  {
    return octave_uint16(pRTreeC->height());
  }

  octave_value identical(RTreeOct *other)
  {
    return pRTreeC->identical(other->pRTreeC);
  }

  void json_write(octave_value fd_ov)
  {
    int fd = fd_ov.int_value();
    pRTreeC->json_write(fd);
  }

  octave_value node_size(void)
  {
    return octave_uint64(pRTreeC->node_size());
  }

  octave_value page_size(void)
  {
    return octave_uint64(pRTreeC->page_size());
  }

  void postscript(octave_value fd_ov,
                  RTreeStyleOct *rtso,
                  octave_value axis_ov,
                  octave_value extent_ov,
                  octave_value margin_ov)
  {
    int fd = fd_ov.int_value();
    std::string axis = axis_ov.string_value();
    float extent = extent_ov.float_value();
    float margin = margin_ov.float_value();

    pRTreeC->postscript(fd, rtso->pRTreeStyleC, axis, extent, margin);
  }

  octave_value rect_size(void)
  {
    return octave_uint64(pRTreeC->rect_size());
  }

  octave_value search(octave_value rect_ov,
                      octave::interpreter &interp,
                      octave_value &f,
                      octave_value &context)
  {
    ColumnVector cv = rect_ov.column_vector_value();
    std::size_t n = cv.numel();
    std::vector<rtree_coord_t> rect(n);

    for (std::size_t i = 0 ; i < n ; i++)
      rect.at(i) = cv(i);

    search_context_t cb_context = {
      .interp = &interp,
      .f = f,
      .context = context
    };

    pRTreeC->search(rect, &cb_context);

    return cb_context.context;
  }

  octave_value to_bsrt(void)
  {
    return pRTreeC->to_bsrt();
  }

  octave_value to_csv(void)
  {
    return pRTreeC->to_csv();
  }

  octave_value to_json(void)
  {
    return pRTreeC->to_json();
  }

  octave_value update(octave::interpreter &interp,
                      octave_value &f,
                      octave_value &context)
  {
    update_context_t cb_context = {
      .interp = &interp,
      .f = f,
      .context = context,
      .dim = pRTreeC->dim()
    };
    return octave_int32(pRTreeC->update(&cb_context));
  }

  octave_value unit_sphere_volume(void)
  {
    return pRTreeC->unit_sphere_volume();
  }

  // not part of the API, needed by Octave

  bool is_defined(void) const { return true; }

private:

  std::shared_ptr<RTreeC> pRTreeC;

  DECLARE_OV_BASE_TYPEID_FUNCTIONS_AND_DATA
};

#endif
