#include <octave/oct.h>
#include <octave/parse.h>

#include "RTreeOct.h"
#include "RTreeStyleOct.h"

#include "rtree.h"
#include "rtree/package.h"

#include "keywords.h"

static RTreeOct* rtree_oct(octave_value);
static RTreeStyleOct* rtree_style_oct(octave_value);

DEFMETHOD_DLD(RTree_dispatch, interp, args, nargout, "RTree_dispatch")
{
  std::string method_name = args(0).string_value();

  switch (method_key(method_name))
    {
    case key_add_rect:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value
          id = args(2),
          rect = args(3);
        rto->add_rect(id, rect);
        return octave_value_list();
      }

    case key_branching_factor:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->branching_factor();
        return octave_value_list(val);
      }

    case key_branch_size:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->branch_size();
        return octave_value_list(val);
      }

    case key_bsrt_read:
      {
        octave_value
          fd = args(1),
          val = RTreeOct::bsrt_read(fd);
        return octave_value_list(val);
      }

    case key_bsrt_write:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value fd = args(2);
        rto->bsrt_write(fd);
        return octave_value_list();
      }

    case key_bugreport:
      {
        octave_value val = RTreeOct::bugreport();
        return octave_value_list(val);
      }

    case key_clone:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->clone();
        return octave_value_list(val);
      }

    case key_constructor:
      {
        octave_value
          dim = args(1),
          split = args(2),
          node_page = args(3),
          val = RTreeOct::constructor(dim, split, node_page);
        return octave_value_list(val);
      }

    case key_csv_read:
      {
        octave_value
          fd = args(1),
          dim = args(2),
          split = args(3),
          node_page = args(4),
          val = RTreeOct::csv_read(fd, dim, split, node_page);
        return octave_value_list(val);
      }

    case key_csv_write:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value fd = args(2);
        rto->csv_write(fd);
        return octave_value_list();
      }

    case key_dim:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->dim();
        return octave_value_list(val);
      }

    case key_empty:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->empty();
        return octave_value_list(val);
      }

    case key_envelope:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->envelope();
        return octave_value_list(val);
      }

    case key_from_bsrt:
      {
        octave_value
          bsrt = args(1),
          val = RTreeOct::from_bsrt(bsrt);
        return octave_value_list(val);
      }

    case key_from_json:
      {
        octave_value
          json = args(1),
          val = RTreeOct::from_json(json);
        return octave_value_list(val);
      }

    case key_height:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->height();
        return octave_value_list(val);
      }

    case key_identical:
      {
        RTreeOct
          *rto1 = rtree_oct(args(1)),
          *rto2 = rtree_oct(args(2));
        octave_value val = rto1->identical(rto2);
        return octave_value_list(val);
      }

    case key_json_read:
      {
        octave_value
          fd = args(1),
          val = RTreeOct::json_read(fd);
        return octave_value_list(val);
      }

    case key_json_write:
      {
        RTreeOct *rto = rtree_oct(args(1));
        int fd = args(2).int_value();
        rto->json_write(fd);
        return octave_value_list();
      }

    case key_name:
      {
        octave_value val = RTreeOct::name();
        return octave_value_list(val);
      }

    case key_node_size:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->node_size();
        return octave_value_list(val);
      }

    case key_page_size:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->page_size();
        return octave_value_list(val);
      }

    case key_postscript:
      {
        RTreeOct *rto = rtree_oct(args(1));
        RTreeStyleOct *rtso = rtree_style_oct(args(3));
        octave_value
          fd = args(2),
          axis = args(4),
          extent = args(5),
          margin = args(6);
        rto->postscript(fd, rtso, axis, extent, margin);
        return octave_value_list();
      }

    case key_rect_size:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->rect_size();
        return octave_value_list(val);
      }

    case key_search:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value
          rect = args(2),
          f = args(3),
          context = args(4),
          val = rto->search(rect, interp, f, context);
        return octave_value_list(val);
      }

    case key_size:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->bytes();
        return octave_value_list(val);
      }

    case key_to_bsrt:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->to_bsrt();
        return octave_value_list(val);
      }

    case key_to_csv:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->to_csv();
        return octave_value_list(val);
      }

    case key_to_json:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->to_json();
        return octave_value_list(val);
      }

    case key_unit_sphere_volume:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value val = rto->unit_sphere_volume();
        return octave_value_list(val);
      }

    case key_update:
      {
        RTreeOct *rto = rtree_oct(args(1));
        octave_value
          f = args(2),
          context = args(3),
          val = rto->update(interp, f, context);
        return octave_value_list(val);
      }

    case key_url:
      {
        octave_value val = RTreeOct::url();
        return octave_value_list(val);
      }

    case key_version:
      {
        octave_value val = RTreeOct::version();
        return octave_value_list(val);
      }

    default:
      error_with_id("RTree:BUG", "unknown method: %s",
                    method_name.c_str());
      return octave_value_list();
    }
}

// helper functions

static RTreeOct* rtree_oct(octave_value arg)
{
  octave_base_value *object = arg.internal_rep();
  RTreeOct *rtree = dynamic_cast <RTreeOct*>(object);

  if (! rtree)
    error_with_id("RTree:not-initialised",
                  "RTree instance not initialised");

  return rtree;
}

static RTreeStyleOct* rtree_style_oct(octave_value arg)
{
  octave_base_value *object = arg.internal_rep();
  RTreeStyleOct *rtree_style = dynamic_cast <RTreeStyleOct*>(object);

  if (! rtree_style)
    error_with_id("RTreeStyle:not-initialised",
                  "RTreeStyle instance not initialised");

  return rtree_style;
}
