#ifndef GENERIC_H
#define GENERIC_H

#include "rtree.h"

#include <octave/error.h>

#include <stdio.h>
#include <unistd.h>

#include <string>

// simple mechanism to run cleanup on scope exit

template <typename F>
struct FinalAction {
  FinalAction(F f) : clean_{f} {}
  ~FinalAction() { clean_(); }
private:
  F clean_;
};

template <typename F>
FinalAction<F> finally(F f) { return FinalAction<F>(f); }

#define FCLOSE_ON_EXIT(st) auto close_stream = finally([st] { fclose(st); })

// use a function which writes a type R to FILE* to create a function
// which writes a type R to std::string

template <typename R>
std::string to_generic(int f(const R*, FILE*), const R *r)
{
  errno = 0;
  FILE *st;

  if ((st = tmpfile()) == NULL)
    error_with_id("RTree:ERRNO", "tmpfile: %s", strerror(errno));

  FCLOSE_ON_EXIT(st);

  if (f(r, st) != 0)
    error_with_id("RTree:ERRNO", "to_generic: %s", strerror(errno));

  if (fseek(st, 0L, SEEK_END) != 0)
    error_with_id("RTree:ERRNO", "fseek: %s", strerror(errno));

  long len;

  if ((len = ftell(st)) < 0)
    error_with_id("RTree:ERRNO", "ftell: %s", strerror(errno));

  rewind(st);

  char buffer[len];
  size_t nread = fread(buffer, 1, len, st);

  if (nread != len)
    error_with_id("RTree:BUG", "fread : %zi/%li", nread, len);

  std::string result;
  result.reserve(len);
  result.append(buffer, len);

  return result;
}

// use a function which reads a type R from FILE* to create a function
// which reads a type R to std::string

template <class T, typename R>
T* from_generic(R* f(FILE*), std::string src)
{
  FILE *st;

  if ((st = tmpfile()) == NULL)
    error_with_id("RTree:ERRNO",
                  "tmpfile: %s", strerror(errno));

  FCLOSE_ON_EXIT(st);

  if (fwrite(src.data(), 1, src.size(), st) != src.size())
    error_with_id("RTree:ERRNO",
                  "fwrite: %s", strerror(errno));

  if (fflush(st) == EOF)
    error_with_id("RTree:ERRNO",
                  "fflush: %s", strerror(errno));

  rewind(st);

  R *r = f(st);

  if (r == NULL)
    {
      if (errno)
        error_with_id("RTree:ERRNO", "%s", strerror(errno));
      else
        error_with_id("RTree:librtree", "parse error");
    }

  return new T(r);
}

// use a function which writes a type R to FILE* to create a function
// which writes a type R to POSIX file-descriptor, variadic

template <typename R, typename ... V>
void generic_write(int fd, std::string mode,
                   std::function<int (const R*, FILE*, V...)>&& f,
                   const R *r, V ... args)
{
  int fd2 = dup(fd);

  if (fd2 == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, mode.c_str())) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  FCLOSE_ON_EXIT(st);

  int err;

  if ((err = f(r, st, args ...)) != 0)
    error_with_id("RTree:librtree", "%s", rtree_strerror(err));
}

template <typename R, typename ... V>
void generic_write(int fd, std::string mode,
                   int f(const R*, FILE*, V...),
                   const R *r, V ... args)
{
  return
    generic_write(fd, mode,
                  std::function<int (const R*, FILE*, V...)>(f),
                  r, args ...);
}

// use a function which reads a type R from FILE* to create a function
// which reads a type R from POSIX file-descriptor, variadic

template <class T, typename R, typename ... V>
T* generic_read(int fd, std::string mode,
                std::function<R* (FILE*, V...)>&& f,
                V ... args)
{
  int fd2;

  if ((fd2 = dup(fd)) == -1)
    error_with_id("RTree:ERRNO", "dup: %s", strerror(errno));

  FILE *st;

  if ((st = fdopen(fd2, mode.c_str())) == NULL)
    error_with_id("RTree:ERRNO", "fdopen: %s", strerror(errno));

  FCLOSE_ON_EXIT(st);

  errno = 0;

  R *r = f(st, args ...);

  if (r == NULL)
    {
      if (errno)
        error_with_id("RTree:ERRNO", "%s", strerror(errno));
      else
        error_with_id("RTree:librtree", "read returns NULL");
    }

  return new T(r);
}

template <class T, typename R, typename ... V>
T* generic_read(int fd, std::string mode,
                R* f(FILE*, V...),
                V ... args)
{
  return
    generic_read<T>(fd, mode,
                    std::function<R* (FILE*, V...)>(f),
                    args ...);
}

#endif
